import { LiquidityPoolJsonInfo } from '@raydium-io/raydium-sdk';
import { Connection, Keypair, Transaction, SystemProgram, LAMPORTS_PER_SOL, TransactionMessage, VersionedTransaction } from '@solana/web3.js';
import { transfer, getAssociatedTokenAddress, createAssociatedTokenAccountInstruction, createTransferInstruction } from "@solana/spl-token";
import RaydiumSwap from '../server/RaydiumSwap';
import { convertKeypair, convertNumber, convertPubkey } from '../utils';

export class SolanaSwapClient {
  rpc: string = 'https://api.mainnet-beta.solana.com';
  connection: Connection;
  poolKeys: LiquidityPoolJsonInfo[];

  constructor(poolKeys: LiquidityPoolJsonInfo[], devnet: boolean = false) {
    this.rpc = devnet ? 'https://api.devnet.solana.com' : 'https://api.mainnet-beta.solana.com';
    this.connection = new Connection(this.rpc);
    if (!poolKeys || poolKeys.length === 0) {
      throw new Error('Pool keys not loaded');
    }
    this.poolKeys = poolKeys;

    if (!this.poolKeys || this.poolKeys.length === 0) {
      throw new Error('Pool keys not loaded');
    }
  }

  public async swap(
    fromPrivKeyBs58: string,
    tokenIn: string,
    tokenOut: string,
    amountIn: string,
    slippage: string = "5", // 5%
    priceImpact: string | null = null,
    maxRetries: number = 20,
  ): Promise<{
    txid: string,
    amountOut: number,
  }> {
    const owner = convertKeypair(fromPrivKeyBs58);
    const amountInSols = convertNumber(amountIn);

    const raydiumSwap = new RaydiumSwap(this.rpc, owner, this.poolKeys);

    const poolInfo = raydiumSwap.findPoolInfoForTokens(tokenIn, tokenOut);
    if (!poolInfo) {
      throw new Error("Pool info not found");
    }

    const transaction = await raydiumSwap.getSwapTransaction(
      tokenOut,
      amountInSols,
      poolInfo,
      slippage,
      priceImpact,
    );

    const txid = await raydiumSwap.sendLegacyTransaction(
      transaction as Transaction,
      maxRetries
    );

    const latestBlockhash = await this.connection.getLatestBlockhash();
    await this.connection.confirmTransaction({
      signature: txid,
      blockhash: latestBlockhash.blockhash,
      lastValidBlockHeight: latestBlockhash.lastValidBlockHeight,
    }, "confirmed");

    const tx = await this.connection.getTransaction(txid, { "commitment": "confirmed", "maxSupportedTransactionVersion": 1 });
    const preTokenOutBalance = tx.meta.preTokenBalances.find(obj => obj.mint === tokenOut && obj.owner === owner.publicKey.toBase58());
    const postTokenOutBalance = tx.meta.postTokenBalances.find(obj => obj.mint === tokenOut && obj.owner === owner.publicKey.toBase58());
    const amountOut = postTokenOutBalance.uiTokenAmount.uiAmount - preTokenOutBalance.uiTokenAmount.uiAmount;

    return {
      txid,
      amountOut
    };
  }

  public async calcAmountOut(
    fromPrivKeyBs58: string,
    tokenIn: string,
    tokenOut: string,
    amountIn: string,
    slippage: string = "5", // 5%
    priceImpact: string | null = null,
  ): Promise<{
    amountIn: number;
    amountOut: number;
  }> {
    const owner = convertKeypair(fromPrivKeyBs58);
    const amountInSols = convertNumber(amountIn);

    const raydiumSwap = new RaydiumSwap(this.rpc, owner, this.poolKeys);

    const poolInfo = raydiumSwap.findPoolInfoForTokens(tokenIn, tokenOut);
    if (!poolInfo) {
      throw new Error("Pool info not found");
    }

    const directionIn = poolInfo.quoteMint.toString() == tokenOut;
    const amountOut = await raydiumSwap.calcAmountOut(
      poolInfo,
      amountInSols,
      directionIn,
      slippage,
      priceImpact,
    );

    const amountOutParsed =
      amountOut.amountOut.raw.toNumber() /
      10 ** amountOut.amountOut.currency.decimals;

    return {
      amountIn: amountInSols,
      amountOut: amountOutParsed,
    };
  }
}

export async function transferSpl(
  connection: Connection,
  fromPrivKeyBs58: string,
  token: string,
  amount: string,
  receiver: string,
): Promise<string> {
  const to = convertPubkey(receiver);
  const from = convertKeypair(fromPrivKeyBs58);
  const amountInSols = convertNumber(amount);

  const fromAssociatedTokenAccount = await getAssociatedTokenAddress(
    convertPubkey(token),
    from.publicKey,
  );
  const toAssociatedTokenAccount = await getAssociatedTokenAddress(
    convertPubkey(token),
    to,
  );

  const tx = new Transaction();

  try {
    const toAccountInfo = await connection.getAccountInfo(toAssociatedTokenAccount);
    if (!toAccountInfo) {
      tx.add(
        createAssociatedTokenAccountInstruction(
          from.publicKey,
          toAssociatedTokenAccount,
          to,
          convertPubkey(token)
        )
      );
    }
  } catch (e) {
    tx.add(
      createAssociatedTokenAccountInstruction(
        from.publicKey,
        toAssociatedTokenAccount,
        to,
        convertPubkey(token)
      )
    );
  }

  tx.add(
    createTransferInstruction(
      fromAssociatedTokenAccount,
      toAssociatedTokenAccount,
      from.publicKey,
      amountInSols * LAMPORTS_PER_SOL,
    )
  );

  tx.recentBlockhash = (await connection.getLatestBlockhash()).blockhash;
  tx.feePayer = from.publicKey;
  const recoverTx = Transaction.from(tx.serialize({ requireAllSignatures: false }));

  // transaction sign
  recoverTx.sign(from);

  const signature = await connection.sendRawTransaction(recoverTx.serialize());
  return signature;
}

export async function transferSols(
  connection: Connection,
  fromPrivKeyBs58: string,
  amount: string,
  receiver: string,
): Promise<string> {
  const to = convertPubkey(receiver);
  const from = convertKeypair(fromPrivKeyBs58);
  const amountInSols = convertNumber(amount);

  const transferInstruction = [
    SystemProgram.transfer({
      fromPubkey: from.publicKey,
      toPubkey: to,
      lamports: amountInSols * LAMPORTS_PER_SOL
    }),
  ];

  let { blockhash } = await connection.getLatestBlockhash();

  const message = new TransactionMessage({
    payerKey: from.publicKey,
    recentBlockhash: blockhash,
    instructions: transferInstruction,
  }).compileToV0Message();

  const transaction = new VersionedTransaction(message);

  transaction.sign([from]);

  const signature = await connection.sendRawTransaction(transaction.serialize());
  return signature;
}


